package be.iminds.ilabt.jfed.log_cache;

import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetails;
import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetailsFactory;
import be.iminds.ilabt.jfed.log.ApiCallDetails;
import be.iminds.ilabt.jfed.log.ApiCallIdGenerator;
import be.iminds.ilabt.jfed.util.library.JFedUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;

import javax.inject.Inject;
import java.io.File;
import java.io.IOException;
import java.nio.file.*;
import java.nio.file.attribute.BasicFileAttributes;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.List;

/**
 * User: twalcari
 * Date: 4/15/2014
 * Time: 15:56
 */
public class DiskApiCallDetailsCache extends ApiCallDetailsCache {

    private static final Logger LOG = LoggerFactory.getLogger(DiskApiCallDetailsCache.class);
    private static final Marker CLEAR_MARKER = MarkerFactory.getMarker("CLEAR");
    private static final Marker LOAD_MARKER = MarkerFactory.getMarker("LOAD");
    private static final Marker REDUCE_MARKER = MarkerFactory.getMarker("REDUCE");
    private static final Marker FLUSH_MARKER = MarkerFactory.getMarker("REDUCE");
    private final Path callDir;
    private final SerializableApiCallDetailsFactory serializableApiCallDetailsFactory;

    @Inject
    DiskApiCallDetailsCache(SerializableApiCallDetailsFactory serializableApiCallDetailsFactory) {
        this.serializableApiCallDetailsFactory = serializableApiCallDetailsFactory;
        Path callDir;

        try {
            callDir = JFedUtils.getUserDataDirectoryPath().resolve("jFedCalls");
            if (!Files.exists(callDir)) {
                Files.createDirectory(callDir);
            }

            //was: File callDir = Files.createTempDirectory("jFedCalls");
            //     IOUtils.deleteDirAtExit(callDir.toFile());
            LOG.debug("Created a tempdir for logging calls at {}", callDir);
        } catch (IOException e) {
            LOG.error("Could not create temporary directory to store jFed calls history!", e);
            callDir = null;
        }
        this.callDir = callDir;
    }

    @Override
    protected SerializableApiCallDetails getFullApiDetails(ApiCallDetailsRef ref) {
        Path path = callDir.resolve(Integer.toString(ref.getId()));

        try {
            byte[] compressedDetails = Files.readAllBytes(path);
            return decompressFullApiDetails(compressedDetails);
        } catch (IOException e) {
            LOG.error("Could not read ApiCallDetails with ref " + ref);
            return null;
        }
    }

    @Override
    public void clear() {
        try {
            Files.walkFileTree(callDir, new SimpleFileVisitor<Path>() {
                @Override
                public FileVisitResult preVisitDirectory(Path dir, BasicFileAttributes attrs) {
                    LOG.debug(CLEAR_MARKER, "Skipping {}", dir.toString());
                    return FileVisitResult.CONTINUE;
                }

                @Override
                public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
                    try {
                        Files.delete(file);
                        LOG.debug(CLEAR_MARKER, "Successfully deleted file {}", file.toString());
                    } catch (IOException ex) {
                        LOG.warn(CLEAR_MARKER, "Could not delete file {}", file.toString());
                        throw ex;
                    }


                    return FileVisitResult.CONTINUE;
                }

                @Override
                public FileVisitResult visitFileFailed(Path file, IOException exc) throws IOException {
                    try {
                        Files.delete(file);
                        LOG.debug("Successfully deleted FAILING file {}", file.toString());
                    } catch (IOException ex) {
                        LOG.warn("Could not delete FAILING file {}", file.toString());
                        throw ex;
                    }

                    return FileVisitResult.CONTINUE;
                }

                @Override
                public FileVisitResult postVisitDirectory(Path dir, IOException exc) {
                    return FileVisitResult.CONTINUE;
                }
            });
        } catch (IOException e) {
            LOG.error("An error occurred while clearing the calls cache", e);
        }
    }


    @Override
    public ApiCallDetailsRef registerApiCallDetails(ApiCallDetails apiCallDetails) {
        ApiCallDetailsRef ref = new ApiCallDetailsRef(apiCallDetails);

        Path path = callDir.resolve(Integer.toString(ref.getId()));

        if (!path.toFile().exists()) {
            try {
                Path newFile =
                        Files.write(path, compressFullApiDetails(
                                serializableApiCallDetailsFactory.createFromApiCallDetails(apiCallDetails),
                                false
                        ));
//                path.toFile().deleteOnExit();
            } catch (IOException e) {
                LOG.error("Error while writing ApiCallDetails to disk", e);
            }
        } else {
            LOG.warn("Detected that ApiCallDetails #{} already exists on disk. Ignoring new ApiCallDetails", apiCallDetails.getId());
        }

        return ref;
    }

    @Override
    public ApiCallDetailsRef registerApiCallDetails(SerializableApiCallDetails apiCallDetails) {
        //first create a copy of the received SerializableApiCallDetails
        SerializableApiCallDetails copy = new SerializableApiCallDetails(apiCallDetails);

        ApiCallDetailsRef ref = new ApiCallDetailsRef(copy);

        Path path = callDir.resolve(Integer.toString(ref.getId()));

        if (!path.toFile().exists()) {
            try {
                Path newFile =
                        Files.write(path, compressFullApiDetails(new SerializableApiCallDetails(apiCallDetails), false));
//                path.toFile().deleteOnExit();
            } catch (IOException e) {
                LOG.error("Error while writing ApiCallDetails to disk", e);
            }
        } else {
            LOG.warn("Detected that ApiCallDetails #{} already exists on disk. Ignoring new ApiCallDetails", apiCallDetails.getId());
        }


        return ref;
    }

    /* @returns null if file is not a call file. */
    private static Integer getIdFromFile(Path file) {
        try {
            return Integer.parseInt(file.getFileName().toString());
        } catch (NumberFormatException e) {
            return null;
        }
    }

    public List<ApiCallDetailsRef> loadExisting() {
        //load (IDs) of all existing call logs
        List<ApiCallDetailsRef> res = new ArrayList<>();
        try {
            Files.walkFileTree(callDir, new SimpleFileVisitor<>() {
                @Override
                public FileVisitResult preVisitDirectory(Path dir, BasicFileAttributes attrs) {
                    return FileVisitResult.CONTINUE;
                }


                @Override
                public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
                    Integer id = getIdFromFile(file);
                    if (id != null) {
                        res.add(new ApiCallDetailsRef(id));
                        LOG.debug(LOAD_MARKER, "Successfully loaded file {}", file.toString());
                    } else {
                        LOG.debug(LOAD_MARKER, "Not loaded: not a call file {}", file.toString());
                    }

                    return FileVisitResult.CONTINUE;
                }

                @Override
                public FileVisitResult visitFileFailed(Path file, IOException exc) throws IOException {
                    LOG.debug(LOAD_MARKER, "Ignoring FAILED file {}", file.toString());

                    return FileVisitResult.CONTINUE;
                }

                @Override
                public FileVisitResult postVisitDirectory(Path dir, IOException exc) {
                    return FileVisitResult.CONTINUE;
                }
            });
        } catch (IOException e) {
            LOG.error("An error occurred while loading the calls cache", e);
        }
        LOG.info("Loaded previous disk call cache: "+res.size()+" calls recovered");
        return res;
    }

    /**
     * remove files that are:
     *   - More than 1000 IDs away from "head" ID
     *   - Older than 5 days
     * */
    private boolean isOld(Path path) {
        Integer id = getIdFromFile(path);
        if (id != null && Math.abs(ApiCallIdGenerator.peekId() - id) > 1000)
            return true;

        Instant daysAgo = Instant.now().minus(5, ChronoUnit.DAYS);
        try {
            BasicFileAttributes attr = Files.readAttributes(path, BasicFileAttributes.class);
            if (attr.creationTime().toInstant().isBefore(daysAgo))
                return true;
            if (attr.lastModifiedTime().toInstant().isBefore(daysAgo))
                return true;
        } catch (IOException e) {
            LOG.error(FLUSH_MARKER, "Error while checking if "+path+" is old. Will assume it is.", e);
            return true;
        }

        return false;
    }

    public void flushOld() {
        //find and remove too old call logs
        try {
            Files.walkFileTree(callDir, new SimpleFileVisitor<>() {
                @Override
                public FileVisitResult preVisitDirectory(Path dir, BasicFileAttributes attrs) {
                    return FileVisitResult.CONTINUE;
                }


                @Override
                public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
                    Integer id = getIdFromFile(file);
                    if (id != null && isOld(file)) {
                        try {
                            Files.delete(file);
                        } catch (IOException e) {
                            LOG.error(FLUSH_MARKER, "Failed to delete {}", file.toString());
                        }
                    }

                    return FileVisitResult.CONTINUE;
                }

                @Override
                public FileVisitResult visitFileFailed(Path file, IOException exc) throws IOException {
                    try {
                        Files.delete(file);
                    } catch (IOException e) {
                        LOG.error(FLUSH_MARKER, "Failed to failed delete {}", file.toString());
                    }

                    return FileVisitResult.CONTINUE;
                }

                @Override
                public FileVisitResult postVisitDirectory(Path dir, IOException exc) {
                    return FileVisitResult.CONTINUE;
                }
            });
        } catch (IOException e) {
            LOG.error("An error occurred while flusing the calls cache", e);
        }
    }

    private void reduce(Path path) {
        Integer id = getIdFromFile(path);
        assert id != null;
        SerializableApiCallDetails full = getFullApiDetails(new ApiCallDetailsRef(id));
        try {
            Files.write(path, compressFullApiDetails(full, true));
        } catch (IOException e) {
            LOG.error(REDUCE_MARKER, "Failed to reduce {}", path.toString());
        } finally {
            //bit of a hack: force garbage collect loaded call now
            full = null;
            System.gc();
        }
    }

    public void reduceOld() {
        //load all existing call logs, and remove large data, then save them again.

        try {
            Files.walkFileTree(callDir, new SimpleFileVisitor<>() {
                @Override
                public FileVisitResult preVisitDirectory(Path dir, BasicFileAttributes attrs) {
                    return FileVisitResult.CONTINUE;
                }


                @Override
                public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
                    Integer id = getIdFromFile(file);
                    if (id != null && Files.size(file) > LARGE_FILE_SIZE) {
                        reduce(file);
                    }

                    return FileVisitResult.CONTINUE;
                }

                @Override
                public FileVisitResult visitFileFailed(Path file, IOException exc) throws IOException {
                    LOG.debug(REDUCE_MARKER, "Ignoring FAILED file {}", file.toString());

                    return FileVisitResult.CONTINUE;
                }

                @Override
                public FileVisitResult postVisitDirectory(Path dir, IOException exc) {
                    return FileVisitResult.CONTINUE;
                }
            });
        } catch (IOException e) {
            LOG.error("An error occurred while flusing the calls cache", e);
        }
    }
}
